#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <string.h>
#include "MPU6050.h"
#include "i2c_master.h"

#define PI 3.14159265
#define RADIANS_TO_DEGREE 180.0 / PI
#define Kp	40
#define Kd  0.05
#define Ki  40
#define sampleTime  0.005
#define targetAngle -2.5
#define MAX_POWER	255
#define MIN_POWER	100
#define ANGLE_ERROR_ACCEPTED_MAX 3
#define ANGLE_ERROR_ACCEPTED_MIN -3


int16_t accX, accY, accZ, gyroX, gyroY, gyroZ;
volatile int gyroRate, motorPower = 0;
volatile float accAngle, gyroAngle, currentAngle, prevAngle=0, error, errorSum=0;

long map_number(long x, long in_min, long in_max, long out_min, long out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

float constrain_number(float x, float a, float b) {
	if (x >= a && x <= b)
		return x;
	if (x < a)
		return a;
	return b;
}

void MOTORS_forward(int motorSpeed) {
	if (motorSpeed > MAX_POWER)
		motorSpeed = MAX_POWER;
	if (motorSpeed < MIN_POWER)
		motorSpeed = MIN_POWER;

	OCR0A = motorSpeed;
	OCR0B = motorSpeed;

	PORTB &= ~(1 << PB0);
	PORTB |= (1 << PB1);
	PORTB |= (1 << PB6);
	PORTB &= ~(1 << PB7);
}

void MOTORS_reverse(int motorSpeed) {
	motorSpeed = fabs(motorSpeed);

	if (motorSpeed > MAX_POWER)
		motorSpeed = MAX_POWER;
	if (motorSpeed < MIN_POWER)
		motorSpeed = MIN_POWER;

	OCR0A = motorSpeed;
	OCR0B = motorSpeed;

	PORTB |= (1 << PB0);
	PORTB &= ~(1 << PB1);
	PORTB &= ~(1 << PB6);
	PORTB |= (1 << PB7);
}

void MOTORS_stop() {
	OCR0A = 0;
	OCR0B = 0;
	PORTB &= ~(1 << PB0);
	PORTB &= ~(1 << PB1);
	PORTB &= ~(1 << PB6);
	PORTB &= ~(1 << PB7);
}

void MOTORS_init()
{
	TCCR0A = 0;
	/* Fast PWM mode
	 * Top at 0xFF
	 */
    TCCR0A |= (3 << WGM00);
    /* Clear on comp OCR0A */
    TCCR0A |= (1 << COM0A1);
    /* Clear on comp OCR0B */
    TCCR0A |= (1 << COM0B1);
    /* prescaler 1*/
    TCCR0B |= (1 << CS00);
}

void MOTORS_setup_pin() {
	/* LEFT_MOTOR PWM PIN*/
	DDRB |= (1 << PB3);
	PORTB &= ~(1 << PB3);

	/* LEFT_MOTOR DIRECTION PINS */
	DDRB |= (1 << PB0);
	PORTB |= (1 << PB0);
	DDRB |= (1 << PB1);
	PORTB &= ~(1 << PB1);

	/* RIGHT_MOTOR PWM PIN*/
	DDRB |= (1 << PB4);
	PORTB &= ~(1 << PB4);

	/* RIGHT_MOTOR DIRECTION PINS */
	DDRB |= (1 << PB6);
	PORTB &= ~(1 << PB6);
	DDRB |= (1 << PB7);
	PORTB |= (1 << PB7);
}

void ROBOT_setup() {
	// set the motor control and PWM pins to output mode
	MOTORS_setup_pin();
	MOTORS_init();

	// initialize the MPU6050
	I2C_init();
	mpu6050_init();
}

void loop() {
	// read acceleration and gyroscope values
	mpu6050_getRawData(&accX, &accY, &accZ, &gyroX, &gyroY, &gyroZ);
 	float angle = (float) accY / (float) accZ;
  	accAngle = atan(angle);
  	accAngle *= RADIANS_TO_DEGREE;
 	gyroRate = map_number(gyroX, -32768, 32767, -250, 250);
	gyroAngle = (float) gyroRate * sampleTime;

	// apply the complementary filter
  	currentAngle = 0.90 * (prevAngle + gyroAngle) + 0.10 * accAngle;

	if (currentAngle >= ANGLE_ERROR_ACCEPTED_MAX || currentAngle <= ANGLE_ERROR_ACCEPTED_MIN) {
		error = currentAngle - targetAngle;
		errorSum = errorSum + error;
		errorSum = constrain_number(errorSum, -300, 300);

		//calculate output from P, I and D values
		motorPower = Kp * error + Ki * errorSum * sampleTime - Kd * (currentAngle - prevAngle) / sampleTime;
		if (motorPower > 0)
			MOTORS_forward(motorPower);
		else if (motorPower < 0)
			MOTORS_reverse(motorPower);
		else
			MOTORS_stop();
	} else {
		MOTORS_stop();
	}
	prevAngle = currentAngle;
}

int main(void) {
	ROBOT_setup();
	while (1) {
		_delay_ms(1);
		loop();
	}
	return 0;
}
